SystemOrganization addCategory: #'SimplePersistence-Core'!

!Class methodsFor: '*SimplePersistence-Core' stamp: 'SeanDeNigris 12/8/2011 00:28'!
persist: aSymbol 

	self persister persist: aSymbol in: self.! !

!Class methodsFor: '*SimplePersistence-Core' stamp: 'SeanDeNigris 2/1/2013 14:37'!
persister

	^ SpFileDatabase.! !

!DataStream classSide methodsFor: '*SimplePersistence-Core' stamp: 'SeanDeNigris 5/29/2011 12:54'!
detectFile: aBlock do: anotherBlock

	^aBlock value
		ifNil: [nil]
      	ifNotNil: [:file| [anotherBlock value: file] ensure: [file close]]! !

!DataStream classSide methodsFor: '*SimplePersistence-Core' stamp: 'SeanDeNigris 5/29/2011 12:54'!
newFileNamed: fileName do: aBlock
	"Returns the result of aBlock."
	
	^ self detectFile: [ self newFileNamed: fileName ] do: aBlock! !

!DataStream classSide methodsFor: '*SimplePersistence-Core' stamp: 'SeanDeNigris 5/29/2011 12:53'!
readOnlyFileNamed: aName do: aBlock 
    | file |
    file := self readOnlyFileNamed: aName.
    ^ file 
        ifNil: [ nil ]
        ifNotNil: [ [ aBlock value: file ] ensure: [ file close ] ]! !

!Object methodsFor: '*SimplePersistence-Core' stamp: 'SeanDeNigris 12/7/2011 23:50'!
persistIn: aClass as: aSymbol

	aClass persist: aSymbol.! !

Object subclass: #PersistenceBehavior
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SimplePersistence-Core'!

PersistenceBehavior class
	instanceVariableNames: 'persistedClass persistedInstance'!

PersistenceBehavior class
	instanceVariableNames: 'persistedClass persistedInstance'!

!PersistenceBehavior class methodsFor: 'as yet unclassified' stamp: 'SeanDeNigris 12/8/2011 00:39'!
addTo: aClass for: aSymbol 

	^ self new addTo: aClass for: aSymbol.! !

!PersistenceBehavior methodsFor: 'as yet unclassified' stamp: 'SeanDeNigris 12/8/2011 00:47'!
addTo: aClass for: aSymbol 

	| genericInstance template source |
	genericInstance := (aClass name first isVowel ifTrue: ['an'] ifFalse: ['a']), aClass name.
	template := 
'restoreFrom: {1}

	{2} := {1}.'.
	source := template format: { genericInstance. aSymbol }.
	aClass class compile: source classified: 'persisting'.! !

Object subclass: #SpFileDatabase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SimplePersistence-Core'!

SpFileDatabase class
	instanceVariableNames: 'lock'!

!SpFileDatabase commentStamp: 'SeanDeNigris 5/29/2011 13:36' prior: 0!
Based on the idea that:
* most applications will not have to scale (i.e. become the next Twitter)
* simply saving the image is slow and error-prone

For the full motivation, see Ramon Leon's blog post at http://onsmalltalk.com/simple-image-based-persistence-in-squeak/.

To give your application persistence:
	1. Subclass SMFileDatabase
	2. On the subclass, implement:
		a. class>>repositories (see method comment).
		b. class>>restoreRepositories: (see method comment).
		
That's it!! Now, whenever you want to save, call class>>saveRepository or class>>takeSnapshot (background save).

To customize:
* Number of backups kept: override class>>defaultHistoryCount
			!

SpFileDatabase class
	instanceVariableNames: 'lock'!

!SpFileDatabase class methodsFor: 'setting up' stamp: 'SeanDeNigris 1/11/2012 20:00'!
addInitializer
	"Call on subclasses to add class initialization"

	self class compile:
'initialize

	self restoreLastBackup.
	self enablePersistence' classified: 'class initialization'.! !

!SpFileDatabase class methodsFor: 'setting up' stamp: 'SeanDeNigris 12/8/2011 00:37'!
addRepositoryAccessorFor: aSymbol of: aClass 

	| template source |
	template := 
'repositories

	^ \{ {1} {2} \}.'.
	source := template format: { aClass name. aSymbol }.
	self class compile: source.! !

!SpFileDatabase class methodsFor: 'setting up' stamp: 'SeanDeNigris 12/8/2011 00:21'!
addRestorationMethodFor: aClass 

	| template source |
	template := 
'restoreRepositories: someRepositories

	{1} restoreFrom: someRepositories first.'.
	source := template format: { aClass name }.
	self class compile: source.! !

!SpFileDatabase class methodsFor: 'private' stamp: 'SeanDeNigris 5/15/2012 23:05'!
backupDirectory

	^ (FileLocator imageDirectory / self name) ensureDirectory.
! !

!SpFileDatabase class methodsFor: 'private' stamp: 'SeanDeNigris 4/11/2012 23:05'!
backupFileForVersion: aNumber

	^ self backupDirectory files detect: [:each | (self versionOf: each) = aNumber ] ifNone: [ nil ].! !

!SpFileDatabase class methodsFor: 'private' stamp: 'SeanDeNigris 2/1/2013 14:12'!
backupVersion: aNumber
	| backup |
	backup := self backupFileForVersion: aNumber.
	backup ifNil: [ ^ nil ].
	^ self serializer materializeFrom: backup fullName.! !

!SpFileDatabase class methodsFor: 'setting up' stamp: 'SeanDeNigris 2/1/2013 14:37'!
createFor: aClass persisting: aSymbol 

	| topLevelCategory dbClassName dbClass |
	topLevelCategory := aClass category copyUpTo: $-.
	dbClassName := (aClass name, 'DB') asSymbol.
	dbClass := SpFileDatabase subclass: dbClassName
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: topLevelCategory, '-Persistence'.
	dbClass
		addInitializer;
		addRepositoryAccessorFor: aSymbol of: aClass;
		addRestorationMethodFor: aClass.! !

!SpFileDatabase class methodsFor: 'private' stamp: 'SeanDeNigris 5/29/2011 13:22'!
defaultHistoryCount
	"How many backups to keep."
	^ 50! !

!SpFileDatabase class methodsFor: 'startup/shutdown' stamp: 'rjl 12/4/2007 18:15'!
enablePersistence
	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self! !

!SpFileDatabase class methodsFor: 'private' stamp: 'SeanDeNigris 12/1/2011 12:55'!
extensionFor: aDirectoryEntryFile
	
	self flag: 'This should be in DirectoryEntryFile, but Pharo 1.4 will have FS and is not accepting fixes to FileDirectory and friends. Squeak 4.2 has DirectoryEntry>>extension, so when 1.4 is released, we can split out platform packages and replace this with the appropriate calls for each fork'.

	^ FileDirectory extensionFor: aDirectoryEntryFile name.! !

!SpFileDatabase class methodsFor: 'private' stamp: 'SeanDeNigris 4/11/2012 23:05'!
lastBackupVersion

	| lastBackupFile |
	lastBackupFile := self backupDirectory files detectMax: [:each | self versionOf: each].
	^ lastBackupFile isNil
		ifTrue: [ 0 ]
		ifFalse: [ self versionOf: lastBackupFile ].
	! !

!SpFileDatabase class methodsFor: 'setting up' stamp: 'SeanDeNigris 12/8/2011 00:05'!
persist: aSymbol in: aClass 

	self createFor: aClass persisting: aSymbol.
	PersistenceBehavior addTo: aClass for: aSymbol.! !

!SpFileDatabase class methodsFor: 'override in subclasses' stamp: 'SeanDeNigris 5/29/2011 13:27'!
repositories
	"Return the root object that needs serialized (e.g. an array containing the root collection of each domain class)."
	self subclassResponsibility! !

!SpFileDatabase class methodsFor: 'restoring' stamp: 'SeanDeNigris 1/11/2012 19:55'!
restoreBackupVersion: aNumber
	(self backupVersion: aNumber) ifNotNilDo: [ :backup | self restoreRepositories: backup ]! !

!SpFileDatabase class methodsFor: 'restoring' stamp: 'SeanDeNigris 1/11/2012 20:33'!
restoreLastBackup

	self restoreBackupVersion: self lastBackupVersion.! !

!SpFileDatabase class methodsFor: 'override in subclasses' stamp: 'SeanDeNigris 5/29/2011 13:29'!
restoreRepositories: someRepositories
	"Put them back  wherever they belong in the image for the application to use them."
	self subclassResponsibility! !

!SpFileDatabase class methodsFor: 'saving' stamp: 'SeanDeNigris 2/1/2013 14:10'!
saveRepository
	| version |
	lock ifNil: [ lock := Semaphore forMutualExclusion ].
	lock critical: [
		self trimBackups.
		version := self lastBackupVersion + 1.
		self serializer serialize: self repositories to: self backupDirectory / self name, version asString ].! !

!SpFileDatabase class methodsFor: 'private' stamp: 'SeanDeNigris 2/1/2013 14:19'!
serializer
	"See SpSerializer class comment"

	^ SpFuelSerializer new.! !

!SpFileDatabase class methodsFor: 'startup/shutdown' stamp: 'rjl 12/4/2007 18:19'!
shutDown: isGoingDown 
	isGoingDown ifTrue: [ self saveRepository ]! !

!SpFileDatabase class methodsFor: 'startup/shutdown' stamp: 'rjl 12/4/2007 18:24'!
startUp: isComingUp 
	isComingUp ifTrue: [ self restoreLastBackup ]! !

!SpFileDatabase class methodsFor: 'restoring' stamp: 'SeanDeNigris 5/29/2011 13:26'!
takeSnapshot
	"Motivation (vs. #restoreLastBackup): if you want a little extra snappiness and you're not worried about making the user wait for the flush to disk"
	[self saveRepository] forkAt: Processor userBackgroundPriority
		named: 'snapshot: ' , self class name! !

!SpFileDatabase class methodsFor: 'private' stamp: 'SeanDeNigris 4/11/2012 23:15'!
trimBackups
	| entries versionsToKeep |
	versionsToKeep := self defaultHistoryCount.
	entries := self backupDirectory files select: [ :e | e base = self name ].
	entries size < versionsToKeep ifTrue: [ ^ self ].
	((entries sortBy: [ :a :b | a extension asInteger < b extension asInteger ]) allButLast: versionsToKeep) 
		do: [ :entry | entry delete ].! !

!SpFileDatabase class methodsFor: 'private' stamp: 'SeanDeNigris 10/29/2011 18:36'!
versionOf: backupFile
	^ (FileDirectory extensionFor: backupFile name) asInteger.! !

Object subclass: #SpSerializer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SimplePersistence-Core'!

!SpSerializer commentStamp: 'SeanDeNigris 2/1/2013 14:19' prior: 0!
I am used by SpFileDatabase to do the actual serialization and materialization. Currently, Fuel is the only supported serializer. The original ReferenceStream serializer has been packaged as SimplePersistence-ReferenceStream just in case a legacy system requires it, but there is no hook to install it, so you'd have to override SpFileDatabase>>serializer.!

!SpSerializer methodsFor: 'protocol' stamp: 'SeanDeNigris 2/1/2013 13:56'!
materializeFrom: aFileReference

	self subclassResponsibility.! !

!SpSerializer methodsFor: 'protocol' stamp: 'SeanDeNigris 2/1/2013 14:02'!
serialize: anObject to: aFileReference

	self subclassResponsibility.! !

!DirectoryEntryFile methodsFor: '*SimplePersistence-Core' stamp: 'SeanDeNigris 12/1/2011 13:03'!
extension
	
	^ FileDirectory extensionFor: self name.! !
